@@ -11,11 +11,12 @@ The algorithm is simple:
 
 `receive_file` launches the command hairgapr that waits for a transfer and exists when a file is received.
 `process_file` read the first bytes of the file
-    - if they match HAIRGAP_MAGIC_NUMBER_INDEX, then this is an index file, with:
-        - the transfer identifier
-        - the previous transfer identifier
-        - the list of following files and their sha256 (in the transfer order)
-    - otherwise, this is the next expected file, as read by the index file
+
+- if they match HAIRGAP_MAGIC_NUMBER_INDEX, then this is an index file, with:
+    - the transfer identifier
+    - the previous transfer identifier
+    - the list of following files and their sha256 (in the transfer order)
+- otherwise, this is the next expected file, as read by the index file
 
 Empty files cannot be sent by hairgap, so they are replaced by the HAIRGAP_MAGIC_NUMBER_EMPTY constant.
 If a new index file is read before the last expected file of the previous index, then we start a new index:
@@ -37,11 +38,15 @@ but can also be run in separate threads for handling large files.
 
 import datetime
 import hashlib
+import io
 import logging
 import os
 import re
+import shlex
 import shutil
 import subprocess
+import tarfile
+import tempfile
 import time
 import uuid
 from queue import Empty, Queue
@@ -53,12 +58,30 @@ from hairgap.constants import (
     HAIRGAP_MAGIC_NUMBER_ESCAPE,
     HAIRGAP_MAGIC_NUMBER_INDEX,
 )
-from hairgap.utils import Config, ensure_dir, now, FILENAME_PATTERN
+from hairgap.utils import Config, FILENAME_PATTERN, ensure_dir, now
 
 logger = logging.getLogger("hairgap")
 
 
 class Receiver:
+    """
+    define the reception process. Can be split into two threads or can be serialize operations when files are small enough.
+    You just have to call the `loop` method to start the reception process.
+    Basically, the algorithm is:
+
+    .. code-block:: python
+
+        while True:
+            receive_file(temporary_filepath)
+            if is_index_file(temporary_filepath):
+                read_list_of_expected_filenames()
+            else:
+                filename = expected_filenames.pop()
+                os.rename(temporary_filepath, filename)
+
+
+    """
+
     available_attributes = set()  # type: Set[str]
 
     def __init__(self, config: Config, threading: bool = False, port: int = None):
@@ -92,6 +115,8 @@ class Receiver:
         # == 0 if there are errors in hairgap (and no file has been created)
         self.current_attributes = {}  # type: Dict[str, str]
         # attributes of the last index
+        self.current_split_status = False
+        # is the last transfer split into chunks?
 
     def receive_file(self, tmp_path) -> Optional[bool]:
         """receive a single file and returns
@@ -115,6 +140,7 @@ class Receiver:
             self.hairgap_subprocess = subprocess.Popen(
                 cmd, stdout=fd, stderr=subprocess.PIPE
             )
+            logger.debug("hairgapr command: %s" % " ".join(cmd))
             __, stderr = self.hairgap_subprocess.communicate()
             fd.flush()
         returncode = self.hairgap_subprocess.returncode
@@ -162,7 +188,7 @@ class Receiver:
         while self.continue_loop:
             try:
                 valid, tmp_abspath = self.process_queue.get(timeout=1)
-                self.process_received_file(tmp_abspath)
+                self.process_received_file(tmp_abspath, valid=valid)
             except Empty:
                 # the timeout is required to quit the thread when self.continue_loop is False
                 continue
@@ -171,16 +197,90 @@ class Receiver:
                 time.sleep(1)
         logger.info("Processing loop exited.")
 
+    @staticmethod
+    def is_gz_file(tmp_abspath: str):
+        if not os.path.isfile(tmp_abspath):
+            return False
+        with open(tmp_abspath, "rb") as fd:
+            header = fd.read(4)
+        return header[:4] == b"\x1f\x8b\x08\x00"
+
     def process_received_file(self, tmp_abspath: str, valid: bool = True):
         """
         process a received file
         the execution time of this method must be small when threading is False (5 seconds between two communications)
         => must be threaded when large files are processed since we compute their sha256.
 
+        :param tmp_abspath: the temporary absolute path
+        :param valid: the file has been correctly received by hairgap
+        :return:
+        """
+        if self.config.use_tar_archives or (
+            self.config.use_tar_archives is None  # auto-detect mode
+            and self.expected_files.empty()
+            and self.is_gz_file(tmp_abspath)
+        ):
+            try:
+                self.process_received_file_tar(tmp_abspath, valid=valid)
+            except Exception as e:
+                logger.exception(e)
+                logger.error("Invalid tar.gz file: %s (removed)" % tmp_abspath)
+                if os.path.isfile(tmp_abspath):
+                    os.remove(tmp_abspath)
+        else:
+            self.process_received_file_no_tar(tmp_abspath, valid=valid)
+
+    def process_received_file_tar(self, tmp_abspath: str, valid: bool = True):
+        """
+        process a tar.gz archive.
+        a single file and a single directory are expected at the root of the received archive
         :param tmp_abspath:
         :param valid:
         :return:
         """
+        if not valid:
+            if os.path.isfile(tmp_abspath):
+                os.remove(tmp_abspath)
+            return
+        with tarfile.open(name=tmp_abspath, mode="r:gz") as tar_fd:
+            index_member = None
+            for member in tar_fd.getmembers():  # type: tarfile.TarInfo
+                if "/" not in member.name and member.isfile():
+                    index_member = member
+                    break
+            if index_member is None:
+                logger.error("index file not found in %s")
+                return
+            # /!\ the index file must be read before extracting other files
+            with tempfile.NamedTemporaryFile() as dst_fd:
+                src_fd = tar_fd.extractfile(index_member)
+                for data in iter(lambda: src_fd.read(8192), b""):
+                    dst_fd.write(data)
+                src_fd.close()
+                dst_fd.flush()
+                self.read_index(dst_fd.name)
+            self.transfer_start()
+            count = 0
+            for member in tar_fd.getmembers():  # type: tarfile.TarInfo
+                if not member.isfile() or member.issym():
+                    continue
+                root, sep, rel_path = member.name.partition("/")
+                if sep != "/":  # the index file => we ignore it
+                    continue
+                self.transfer_file_received(
+                    tmp_abspath,
+                    rel_path,
+                    expected_sha256=None,
+                    actual_sha256=None,
+                    tmp_fd=tar_fd.extractfile(member),
+                )
+                count += 1
+            if count == 0:
+                ensure_dir(self.get_current_transfer_directory(), parent=False)
+            self.transfer_complete()
+        os.remove(tmp_abspath)
+
+    def process_received_file_no_tar(self, tmp_abspath: str, valid: bool = True):
         empty_prefix = HAIRGAP_MAGIC_NUMBER_EMPTY.encode()
         index_prefix = HAIRGAP_MAGIC_NUMBER_INDEX.encode()
         escape_prefix = HAIRGAP_MAGIC_NUMBER_ESCAPE.encode()
@@ -205,6 +305,7 @@ class Receiver:
             self.transfer_start()
             if self.expected_files.empty():
                 # empty transfer => we mark it as complete
+                ensure_dir(self.get_current_transfer_directory(), parent=False)
                 self.transfer_complete()
         elif self.expected_files.empty():
             if valid:
@@ -226,6 +327,10 @@ class Receiver:
             )
             if self.expected_files.empty():
                 # all files of the transfer have been received
+                if self.current_split_status:
+                    self.unsplit_received_files(
+                        self.config, self.get_current_transfer_directory()
+                    )
                 self.transfer_complete()
 
     def transfer_start(self):
@@ -240,17 +345,52 @@ class Receiver:
         """called when all files of a transfer are received.
 
         the execution time of this method must be small when threading is False (5 seconds between two communications)
+        You can read :attr:`current_attributes` to retrieve the attributes defined by the sender (set to `None` by default).
 
         """
         pass
 
     def get_current_transfer_directory(self) -> Optional[str]:
         """return a folder name where all files of a transfer can be moved to.
+
+        The index file has been read and the attributes are set.
         This folder will be automatically created.
         If None, all received files will be deleted.
         """
         raise NotImplementedError
 
+    @staticmethod
+    def unsplit_received_files(config: Config, dir_abspath):
+        names = os.listdir(dir_abspath)
+        if not names:
+            return
+        folder_1 = os.path.join(dir_abspath, str(uuid.uuid4()))
+        folder_2 = os.path.join(dir_abspath, str(uuid.uuid4()))
+        ensure_dir(folder_1, parent=False)
+        ensure_dir(folder_2, parent=False)
+        for name in names:
+            os.rename(os.path.join(dir_abspath, name), os.path.join(folder_1, name))
+        names.sort()
+        cat_cmd = [config.cat] + names
+        tar_cmd = [config.tar, "xz", "-C", folder_2]
+        esc_tar_cmd = [shlex.quote(x) for x in tar_cmd]
+        esc_cat_cmd = [shlex.quote(x) for x in cat_cmd]
+        cmd = "%s | %s" % (" ".join(esc_cat_cmd), " ".join(esc_tar_cmd))
+        p = subprocess.Popen(
+            cmd,
+            shell=True,
+            stdout=subprocess.PIPE,
+            stderr=subprocess.PIPE,
+            stdin=subprocess.PIPE,
+            cwd=folder_1,
+        )
+        p.communicate(b"")
+        names = os.listdir(folder_2)
+        for name in names:
+            os.rename(os.path.join(folder_2, name), os.path.join(dir_abspath, name))
+        shutil.rmtree(folder_1)
+        shutil.rmtree(folder_2)
+
     # noinspection PyMethodMayBeStatic
     def transfer_file_unexpected(self, tmp_abspath: str, prefix: bytes = None):
         """called when an unexpected file has been received. Probably an interrupted transfer…
@@ -268,8 +408,9 @@ class Receiver:
         self,
         tmp_abspath,
         file_relpath,
-        actual_sha256: str = None,
-        expected_sha256: str = None,
+        actual_sha256: Optional[str] = None,
+        expected_sha256: Optional[str] = None,
+        tmp_fd: io.BytesIO = None,
     ):
         """called when a file is received
 
@@ -277,11 +418,26 @@ class Receiver:
 
         :param tmp_abspath: the path of the received file
         :param file_relpath: the destination path of the received file
-        :param actual_sha256: actual SHA256
-        :param expected_sha256: expected SHA256
+        :param actual_sha256: actual SHA256 (not provided in case of tar archives)
+        :param expected_sha256: expected SHA256 (not provided in case of tar archives)
+        :param tmp_fd: provided when tmp_abspath is not given
         :return:
         """
-        if os.path.isfile(tmp_abspath):
+        if tmp_fd:
+            receive_path = self.get_current_transfer_directory()
+            self.transfer_received_count += 1
+            size = 0
+            if receive_path:
+                file_abspath = os.path.join(receive_path, file_relpath)
+                ensure_dir(file_abspath, parent=True)
+                with open(file_abspath, "wb") as dst_fd:
+                    for data in iter(lambda: tmp_fd.read(8192), b""):
+                        dst_fd.write(data)
+                        size += len(data)
+                    tmp_fd.close()
+            else:
+                logger.warning("No receive path defined: ignoring %s." % file_relpath)
+        elif os.path.isfile(tmp_abspath):
             size = os.path.getsize(tmp_abspath)
             self.transfer_received_count += 1
             receive_path = self.get_current_transfer_directory()
@@ -318,8 +474,11 @@ class Receiver:
 
         self.expected_files = Queue()
         expected_count = 0
+        self.current_split_status = False
         with open(index_abspath) as fd:
             for line in fd:
+                if line == "[splitted_content]\n":
+                    self.current_split_status = True
                 matcher = re.match(FILENAME_PATTERN, line)
                 if matcher:
                     self.expected_files.put((matcher.group(1), matcher.group(2)))
